/***************************************************************
This code was written by Franz Franchetti, www.spiral.net --
Copyright (c) 2005, Carnegie Mellon University
All rights reserved.
The code is distributed under a BSD style license
(see http://www.opensource.org/licenses/bsd-license.php)

Redistribution and use in source and binary forms, with or without
modification, are permitted provided that the following conditions are
met:

* Redistributions of source code must retain the above copyright
  notice, reference to Spiral, this list of conditions and the
  following disclaimer.
* Redistributions in binary form must reproduce the above
  copyright notice, this list of conditions and the following
  disclaimer in the documentation and/or other materials provided
  with the distribution.
* Neither the name of Carnegie Mellon University nor the name of its
  contributors may be used to endorse or promote products derived from
  this software without specific prior written permission.

THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
*AS IS* AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
(INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
******************************************************************/

#include <stdio.h>
#include "threads.h"
#include "smp.h"
#include "rdtsc.h"

#define NUM_PROC		2
#define	CHECK_RUNS		1000
#define	MEASURE_RUNS	100000


long volatile data[NUM_PROC];

//	multithreaded worker
double worker(void)
{
	int i, j, tid;
	tsc_counter c0, c1;

	tid = register_thread();

	for (i = 0; i < CHECK_RUNS; i++) {
		data[tid]++;
		barrier(tid, NUM_PROC);
		for (j = 0; j< NUM_PROC; j++)
			if (data[j] != data[tid])
				printf("Error: iter=%d, tid=%d, idx=%d\n", i, tid, j);
		barrier(tid, NUM_PROC);
	}

	RDTSC(c0);
	for (i = 0; i < MEASURE_RUNS; i++)
		barrier(tid, NUM_PROC);
	RDTSC(c1);
	return COUNTER_DIFF(c1, c0, CYCLES)/MEASURE_RUNS;
}


void _worker(void *param) {
	worker();
}


int main()
{
	double cycles;
	int i;
	
	init_threads(NUM_PROC);

	for (i = 0; i++; i < NUM_PROC)
		data[i]=0;

	//	start worker threads
	for (i=1; i<NUM_PROC; i++)
		CREATE_THREAD_NODATA(_worker);

	//	main thread = worker thread 0
	cycles = worker();

	printf("barrier: %d runs, %d cycles/run\n", MEASURE_RUNS, (int)cycles);

	return 0;
}
