/***************************************************************
This code was written by Franz Franchetti
See www.ece.cmu.edu/~franzf and www.spiral.net for details
Copyright (c) 2005, Carnegie Mellon University
All rights reserved.

This code is distributed for evaluation only and may not be 
used for any other purposes. Redistribution without prior consent
by the authors is prohibited.

Redistribution and use in source and binary forms, with or without
modification, are permitted provided that the following conditions are
met:

* Redistributions of source code must retain the above copyright
  notice, reference to Spiral, this list of conditions and the
  following disclaimer.
* Redistributions in binary form must reproduce the above
  copyright notice, this list of conditions and the following
  disclaimer in the documentation and/or other materials provided
  with the distribution.
* Neither the name of Carnegie Mellon University nor the name of its
  contributors may be used to endorse or promote products derived from
  this software without specific prior written permission.

THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
*AS IS* AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
(INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
******************************************************************/

#include "threads.h"
#include "smp.h"

//	global thread library variables
volatile long _current_id = -1;
int _num_threads = 0;
__declspec(thread) long _thread_id;

__declspec(thread) int _ctr2[128];
volatile __declspec(aligned(CACHELINE)) barrier_t bval2[D21*2];

__declspec(thread) int _ctr4[128];
volatile __declspec(aligned(CACHELINE)) barrier_t bval4[D41*4];

__declspec(thread) int _ctrn[128];
volatile int bctr[2] = {0,0}; 


int register_thread() {
#ifdef USE_WIN_THREADS
	HANDLE htid = GetCurrentThread();
#else
	unsigned long mask;
#endif
	_thread_id = _InterlockedIncrement(&_current_id);
	if (_thread_id + 1 == _num_threads) _current_id = -1;
#ifdef USE_WIN_THREADS
	SetThreadAffinityMask(htid, 1 << _thread_id);
//	SetThreadPriority(htid, THREAD_PRIORITY_HIGHEST);
#else
/*	UNTESTED by FF

	Beware that you need the NPTL-version of the pthreads 
	library for this to work (if you do not have it installed, 
	the compiler will complain that it does not know the 
	pthread_setaffinity_np function).

	you have to #include "nptl/pthread.h" not pthread.h, 
	which will probably use the linuxthreads header file which 
	doesn't have the affinity methods. you will also need to do a 
	-I/usr/include/nptl and -L/usr/lib/nptl. be aware you're breaking 
	your program's compatibility with linuxthreads as a result. 

*/

	mask << _thread_id; 
	pthread_setaffinity_np(pthread_self(), sizeof(mask), &mask) <0); 
#endif
	CTR2(_thread_id) = 2;
	CTR4(_thread_id) = 2;
	CTRN(_thread_id) = 0;
	return _thread_id;
}

int get_num_threads() {
	return _num_threads;
}

void init_threads(int nt) { 
	if (nt > MAX_PROC) exit(-1);
	_num_threads = nt;
	ENTRY2(0,0) = 0;
	ENTRY2(0,1) = 0;
	ENTRY2(1,0) = 0;
	ENTRY2(1,1) = 0;
	ENTRY4(0,0) = 0;
	ENTRY4(0,1) = 0;
	ENTRY4(1,0) = 0;
	ENTRY4(1,1) = 0;
	ENTRY4(2,0) = 0;
	ENTRY4(2,1) = 0;
	ENTRY4(3,0) = 0;
	ENTRY4(3,1) = 0;
}

void cleanup_threads() { 
	_current_id = -1;
	_num_threads = 0;
}

