/***************************************************************
This code was written by Franz Franchetti
See www.ece.cmu.edu/~franzf and www.spiral.net for details
Copyright (c) 2005, Carnegie Mellon University
All rights reserved.

This code is distributed for evaluation only and may not be 
used for any other purposes. Redistribution without prior consent
by the authors is prohibited.

Redistribution and use in source and binary forms, with or without
modification, are permitted provided that the following conditions are
met:

* Redistributions of source code must retain the above copyright
  notice, reference to Spiral, this list of conditions and the
  following disclaimer.
* Redistributions in binary form must reproduce the above
  copyright notice, this list of conditions and the following
  disclaimer in the documentation and/or other materials provided
  with the distribution.
* Neither the name of Carnegie Mellon University nor the name of its
  contributors may be used to endorse or promote products derived from
  this software without specific prior written permission.

THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
*AS IS* AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
(INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
******************************************************************/

#ifndef __SMP_H__
#define __SMP_H__

#include <emmintrin.h>

long  __cdecl _InterlockedIncrement(long volatile *Addend);

#define FASTBARRIER
#define	NUMABARRIER
#undef  SAFETY

#ifdef SAFETY
#define FENCE _mm_mfence()
#define PAUSE _mm_pause()
#else
#define FENCE 0
#define PAUSE 0
#endif

//	magic numbers
#define	D21			32
#define	D22			16
#ifdef NUMABARRIER
#define	D41			32
#define	D42			4
#else
#define	D41			32
#define	D42			16
#endif
#define	CACHELINE	32

#define ENTRY2(l,e)	bval2[((l)*D21) + ((e)*D22)]
#define ENTRY4(l,e)	bval4[((l)*D41) + ((e)*D42)]

#define	CTR2(tid)	_ctr2[0]
#define	CTR4(tid)	_ctr4[0]
#define	CTRN(tid)	_ctrn[0]

typedef __int32 barrier_t;

extern __declspec(thread) int _ctr2[128];
extern volatile __declspec(aligned(CACHELINE)) barrier_t bval2[D21*2];

extern __declspec(thread) int _ctr4[128];
extern volatile __declspec(aligned(CACHELINE)) barrier_t bval4[D41*4];

extern __declspec(thread) int _ctrn[128];
extern volatile int bctr[2]; 

void __forceinline barrier(int tid, int nthreads)
{
    int val, blk, lblk, rblk, el, id, ctr2, ctr4, ctrn;
	
	FENCE;
	switch (nthreads) {
#ifdef FASTBARRIER
		case 2:
			ctr2 = CTR2(tid);
			val = (barrier_t)(ctr2 >> 1);
			el = ctr2 & 1;
			blk = tid ^ el;

			ENTRY2(blk, el) = val;
			while (ENTRY2(!blk ,el) != val) 
				PAUSE;

			CTR2(tid) = (ctr2 + 1) & 3;
			break;

		case 4:	
			ctr4 = CTR4(tid);
			val = (barrier_t)(ctr4 >> 1);
			el = ctr4 & 1;
#ifdef NUMABARRIER
			if (!(tid & 1)) {
				blk = (el << 1) ^ tid;
				lblk = tid + 1;
				rblk = blk ^ 2;

				while (ENTRY4(lblk ,el) != val)
					PAUSE;
				ENTRY4(blk, el) = val;
				while (ENTRY4(rblk ,el) != val)
					PAUSE;
			}
			else {
				ENTRY4(tid, el) = val;
				while (((ENTRY4(0, el) != val) || (ENTRY4(2, el) != val)))
					PAUSE;
			}
#else
			blk = tid ^ el;
			ENTRY4(blk, el) = val;
			while ((ENTRY4(blk ^ 1, el) != val) || (ENTRY4(blk ^ 2, el) != val) || (ENTRY4(blk ^ 3, el) != val))
				PAUSE;
#endif
			CTR4(tid) = (ctr4 + 1) & 3;
			break;
#endif
		default:
			CTRN(tid) ^= 1;
			id = CTRN(tid);
			_InterlockedIncrement(&bctr[id]);
			if (!tid)
			{
				while (bctr[id] < nthreads);
				bctr[id] = 0;
			}
			else
				while (bctr[id]);
	}
}


int register_thread(void);
int get_num_threads(void);
void init_threads(int);
void cleanup_threads(void);


#endif
